/*
  Macro     Svn
  Author    Carlo.Hogeveen@xs4all.nl
  Date      25 May 2012
  Version   See the history and the help_text.

  Purpose:
    The fastest possible Subversion browser.

  Why:
    Tortoisesvn is beyond too slow for directories with many files.

  Trade-off:
    A selected file's Subversion-properties are initially not shown.

  Installation:
    The "svn" macro is "just" a shell around the "svn" commandline tool,
    which is [a part of] the actual Subversion tool.
    Therefore the commandline-tool "svn" must be installed and in the PATH
    environment variable, either in general or from where TSE is started.

    Either download the latest version of TortoiseSVN (1.7.7 as I write,
    which also includes and installs "SVN") from
    http://tortoisesvn.net/downloads.html ,

    or download one of the binary packages for "svn" for Windows at the
    bottom of the page of
    http://subversion.apache.org/packages.html .

    As for the macro, just put it in TSE's "mac" directory,
    and compile and execute it.

  Features:
    Browsing the HEAD of a Subversion repository's directories and files,
    view their info, properties, and history, and read or edit [the historic
    version of] a file (without saving to the repository).

  NOT a feature (yet):
    Subversion security: the repository must either be publicly accessible,
    or the username and password must already be locally stored
    by (Tortoise)SVN itself.

  Whishlist:
  MUST
  - Being aksed for a username and password when required by the repository.
  SHOULD
  - Closing a selected historic object should return to the history-list.
  - Being able to browse historic repository directories.
  - In the history-list you shouldn't be able to select the separation lines.
  - Putting the @<revision> before the file-extension to enable
    syntax-hiliting.
  COULD
  - Being able to compare two versions.
  WOULD
  - Make the macro also compileable to a .exe.

  History:
  0.9   25 May 2012
    - Initial version.
  0.9.1  3 Jun 2012
    - If started from the command-line with "-eSvn" and no file was opened
      for editing, then TSE is closed with the macro.
  0.9.2  25 Feb 2013
    - Changed the helptext and a warning to English.
*/

#define LANGUAGE _DEFAULT_


Datadef help_text
  'This is fast Subversion browser, specifically intended for browsing'
  'those really large large directories, for which TortoiseSVN would take'
  'minutes instead of seconds.'
  ''
  'The top of the screen shows which object is currently selected.'
  'A @ followed by a revisionnumber indicates you look'
  'at that historic version (revision) of the object.'
  'Is there no @, then you look at the latest version.'
  ''
  'The bottom of the screen continuously shows possible actions.'
  'A file is opened for read first, and for editing separately.'
  'Editing leaves the browser, and does not mean you can Save the file'
  '(directly) but "only" that you get all the editing options of TSE.'
  '"Props" stands for "Properties": an object' + "'s Subversion-properties."
  ''
  'In a list and when reading a file you can type a string of'
  'characters to quickly go to their first occurrence.'
  'For instance, type "props" now.'
  ''
  'Browse keys: Arrow(Up/Down), PageUp, PageDown, Home, End.'
end help_text

#define STATE_ERROR     0
#define STATE_OK        1
#define STATE_STOPPED   3

string  curr_list         [MAXSTRINGLEN] = ''
string  file_revision     [MAXSTRINGLEN] = ''
string  list_header       [MAXSTRINGLEN] = ''
string  log_file          [MAXSTRINGLEN] = ''
integer log_id                           = 0
string  list_footer       [MAXSTRINGLEN] = ''
string  macro_name        [MAXSTRINGLEN] = ''
string  next_list         [MAXSTRINGLEN] = 'browse'
integer org_id                           = 0
string  selected_file     [MAXSTRINGLEN] = ''
string  selected_property [MAXSTRINGLEN] = ''

Keydef extra_list_keys
  <f1> next_list = 'help'     PushKey(<Enter>)
  <f5> next_list = 'log'      PushKey(<Enter>)
  <f8> next_list = 'info'     PushKey(<Enter>)
  <f9> next_list = 'proplist' PushKey(<Enter>)
end

integer proc get_dos(string cmd)
  integer result = FALSE
  EraseDiskFile(log_file)
  if Dos(cmd + ' > ' + QuotePath(log_file) + ' 2>&1', _START_HIDDEN_)
    GotoBufferId(log_id)
    EmptyBuffer()
    if FileExists(log_file)
      InsertFile(log_file, _DONT_PROMPT_)
      UnMarkBlock()
      EndFile()
      while CurrLine() > 1
      and   CurrLineLen() == 0
          KillLine()
          Up()
      endwhile
      BegFile()
      EraseDiskFile(log_file)
      result = TRUE
    endif
  endif
  return(result)
end get_dos

proc show_dos_error(string text)
  string warning [MAXSTRINGLEN] = text
  BegFile()
  repeat
    warning = warning + Chr(13) + RTrim(GetText(1, MAXSTRINGLEN))
  until not Down()
  Warn(warning)
end show_dos_error

integer proc set_log_file()
  integer org_id = GetBufferId()
  integer result = STATE_ERROR
  string  tmp_dir  [MAXSTRINGLEN] = ''
  if GetEnvStr('tmp') <> ''
    tmp_dir = GetEnvStr('tmp')
  elseif GetEnvStr('temp') <> ''
    tmp_dir = GetEnvStr('temp')
  else
    tmp_dir = 'c:'
  endif
  if tmp_dir[Length(tmp_dir)] <> '\'
    tmp_dir = tmp_dir + '\'
  endif
  if (FileExists(tmp_dir) & _DIRECTORY_)
    log_file = tmp_dir + 'TseSvn.log'
    log_id   = EditFile(log_file, _DONT_PROMPT_)
    if log_id
      EmptyBuffer()
      InsertText('Hello world!')
      if SaveAs(CurrFilename(), _DONT_PROMPT_|_OVERWRITE_)
        result = STATE_OK
      else
        AbandonFile(log_id)
      endif
      GotoBufferId(org_id)
    endif
  endif
  if result <> STATE_OK
    Warn('Error: cannot create a temporary file: ', log_file)
  endif
  return(result)
end set_log_file

integer proc ask_repository(var string repository,
                            var string dir,
                            var string selected)
  integer state                  = STATE_OK
  string  request [MAXSTRINGLEN] = ''
  if  Ask('Subversion repository, directory or selected:',
          request,
          GetFreeHistory(macro_name + ':request'))
    request = Trim(request)
    while SubStr(request, Length(request), 1) == '/'
      request = SubStr(request, 1, Length(request) - 1)
    endwhile
    if request == ''
      state = STATE_STOPPED
    else
      if get_dos('svn info ' + QuotePath(request))
        if lFind('^Repository Root: ', 'gx')
          repository = Trim(GetText(18, MAXSTRINGLEN))
          if lFind('^URL: ', 'gx')
            dir = Trim(GetText(6 + Length(repository), MAXSTRINGLEN))
            if lFind('^Node Kind: {directory}|{selected}', 'gx')
              if lFind('^Node Kind: selected', 'cgx')
                selected = GetToken(dir, '/\', NumTokens(dir, '/\'))
                dir  = SubStr(dir, 1, Length(dir) - Length(selected) - 1)
              else
                selected = ''
              endif
            else
              state = STATE_ERROR
              show_dos_error('Error: no line starting with "Node Kind: file|directory".')
            endif
          endif
        else
          state = STATE_ERROR
          show_dos_error('Error: no line starting with "Repository Root: ".')
        endif
      else
        state = STATE_ERROR
      endif
    endif
  else
    state = STATE_STOPPED
  endif
  return(state)
end ask_repository

proc list_startup()
  UnHook(list_startup)
  case next_list
    when 'browse'
      Enable(extra_list_keys)
  endcase
  ListFooter(list_footer)
end list_startup

proc list_cleanup()
  Disable(extra_list_keys)
end list_cleanup

integer proc browse_repository(string repository, var string dir)
  integer old_msglevel
  integer state                   = STATE_OK
  list_header = repository                                        +
                iif(dir           == '', '', '/' + dir          ) +
                iif(selected_file == '', '', '/' + selected_file) +
                iif(file_revision == '', '', '@' + file_revision)
  list_footer = '{Enter}-Back {Escape}-Back'
  curr_list   = next_list
  case next_list
    when 'browse'
      list_header = repository +
                    iif(dir           == '', '', '/' + dir)
      list_footer = '{F1}-Help {F5}-Hist {F8}-Info {F9}-Props {Enter}-Read {Esc}-Quit'
      get_dos('svn list ' + QuotePath(repository + '/' + dir))
      if lFind('^svn: ', 'gx')
        state = STATE_ERROR
        show_dos_error('Error:')
      else
        MarkLine(1, NumLines())
        old_msglevel = Set(MsgLevel, _WARNINGS_ONLY_)
        ExecMacro('sort -i')
        Set(MsgLevel, old_msglevel)
        UnMarkBlock()
        // Temporarily add a line, because searching 'somthing$' with 'bgx'
        // fails (erroneously!) when the file contains a single line.
        EndFile()
        AddLine('')
        while lFind('/$', 'bgx')
          MarkColumn(CurrLine(), 1, CurrLine(), CurrCol() - 1)
          Copy()
          KillLine()
          BegFile()
          InsertLine('/')
          EndLine()
          Paste()
          UnMarkBlock()
        endwhile
        EndFile()
        KillLine()
        BegFile()
        if dir == ''
          InsertLine('/.')
        else
          InsertLine('/..')
        endif
      endif
      lFind(selected_file, 'g')
    when 'cat'
      list_footer = '{Enter}-Edit {Escape}-Back'
      get_dos('svn cat ' +
              iif(file_revision == '', '', '-r' + file_revision + ' ') +
              QuotePath(repository + '/' + dir + '/' + selected_file))
      if lFind('^svn: ', 'gx')
        state = STATE_ERROR
        show_dos_error('Error:')
      endif
    when 'edit'
      get_dos('svn cat ' +
              iif(file_revision == '', '', '-r' + file_revision + ' ') +
              QuotePath(repository + '/' + dir + '/' + selected_file))
      if lFind('^svn: ', 'gx')
        state = STATE_ERROR
        show_dos_error('Error:')
      else
        MarkLine(1, NumLines())
        Copy()
        org_id = NewFile()
        Paste()
        UnMarkBlock()
        ChangeCurrFilename(list_header, _DONT_EXPAND_)
        state = STATE_STOPPED
      endif
    when 'help'
      EmptyBuffer()
      InsertData(help_text)
      BegFile()
    when 'info'
      get_dos('svn info ' + QuotePath(repository + '/' + dir + '/' + selected_file))
      if lFind('^svn: ', 'gx')
        state = STATE_ERROR
        show_dos_error('Error:')
      endif
    when 'log'
      list_footer = '{Enter}-Read {Esc}-Back'
      get_dos('svn log ' + QuotePath(repository + '/' + dir + '/' + selected_file))
      if lFind('^svn: ', 'gx')
        state = STATE_ERROR
        show_dos_error('Error:')
      else
        while lFind('^$', 'gx')
          KillLine()
        endwhile
      endif
    when 'proplist'
      get_dos('svn proplist -v ' + QuotePath(repository + '/' + dir + '/' + selected_file))
      if lFind('^svn: ', 'gx')
        state = STATE_ERROR
        show_dos_error('Error:')
      else
        lFind(selected_property, 'g')
      endif
    otherwise
      Warn('Error: unknown action (1).')
      state = STATE_ERROR
  endcase
  if state == STATE_OK
    Hook(_LIST_STARTUP_, list_startup)
    Hook(_LIST_CLEANUP_, list_cleanup)
    if list(list_header,
            Max(Max(Length(list_header), Length(list_footer)),
                LongestLineInBuffer()))
      UnHook(list_cleanup)
      case curr_list
        when 'browse'
          selected_file = GetText(1, CurrLineLen())
          if next_list == 'browse'
            if selected_file == '/..'
              selected_file = '/' + SplitPath(dir, _NAME_|_EXT_)
              dir  = SplitPath(dir, _PATH_)
              if SubStr(dir, Length(dir), 1) == '/'
                dir = SubStr(dir, 1, Length(dir) - 1)
              endif
            elseif selected_file == '/.'
              dir = dir
            elseif SubStr(selected_file, 1, 1) == '/'
              if dir == ''
                dir = SubStr(selected_file, 2, MAXSTRINGLEN)
              else
                dir = dir + selected_file
              endif
            else
              next_list     = 'cat'
              file_revision = ''
            endif
          else
            if selected_file == '/..'
              next_list = 'browse'
            endif
          endif
        when 'cat'
          next_list = 'edit'
        when 'help'
          next_list = 'browse'
        when 'info'
          next_list = 'browse'
        when 'log'
          if lFind('^-#$', 'cgx')
            next_list = 'log'
          else
            if not lFind('^r[0-9]# \| ', 'cgx')
              Up()
              if not lFind('^r[0-9]# \| ', 'cgx')
                Down()
              endif
            endif
            if lFind('^r[0-9]# \| ', 'cgx')
              lFind('[0-9]#', 'cgx')
              file_revision = GetFoundText()
              next_list     = 'cat'
            else
              next_list     = 'log'
            endif
          endif
        when 'proplist'
          next_list = 'browse'
        otherwise
          Warn('Error: unknown action (2).')
          state = STATE_ERROR
      endcase
    else
      case curr_list
        when 'browse'
          state = STATE_STOPPED
        when 'cat'
          next_list = 'browse'
        when 'help'
          next_list = 'browse'
        when 'info'
          next_list = 'browse'
        when 'log'
          next_list = 'browse'
        when 'proplist'
          next_list = 'browse'
        otherwise
          Warn('Error: unknown action (3).')
          state = STATE_ERROR
      endcase
    endif
  endif
  return(state)
end browse_repository

proc WhenLoaded()
  macro_name = SplitPath(CurrMacroFilename(), _NAME_)
end WhenLoaded

proc Main()
  string  dir        [MAXSTRINGLEN] = ''
  string  repository [MAXSTRINGLEN] = ''
  integer state                     = STATE_OK
  org_id                            = GetBufferId()
  state = set_log_file()
  while state == STATE_OK
    state = ask_repository(repository, dir, selected_file)
    while state == STATE_OK
      state = browse_repository(repository, dir)
    endwhile
  endwhile
  GotoBufferId(org_id)
  AbandonFile(log_id)
  PurgeMacro(macro_name)
  if  (Lower(Trim(Query(DosCmdLine))) in '-esvn', '-e svn')
  and NumFiles() == 1
  and Pos('unnamed', Lower(CurrFilename())) > 0
    AbandonEditor()
  endif
end Main
